from niaaml_gui.windows.process_window import ProcessWindow
from niaaml_gui.windows.csv_editor_window import CSVEditorWindow
from niaaml_gui.windows import threads

__all__ = ["threads", "ProcessWindow", "CSVEditorWindow"]



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QLineEdit, QHBoxLayout, QVBoxLayout, QFileDialog, QCheckBox
from niaaml_gui.widgets.base_main_widget import BaseMainWidget
from niaaml_gui.windows import ProcessWindow
from niaaml_gui.process_window_data import ProcessWindowData
import qtawesome as qta


class UsePipelineWidget(BaseMainWidget):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        vBoxLayout = QVBoxLayout(self._parent)
        vBoxLayout.setAlignment(Qt.AlignmentFlag.AlignTop)

        selectPPLNFileBar = QHBoxLayout(self._parent)
        selectPPLNFileBar.setSpacing(0)
        fNameLine1 = QLineEdit(self._parent)
        fNameLine1.setObjectName("pplnFile")
        fNameLine1.setPlaceholderText("Select a pipeline file...")
        fNameLine1.setReadOnly(True)
        font = fNameLine1.font()
        font.setPointSize(12)
        fNameLine1.setFont(font)
        selectPPLNFileBar.addWidget(fNameLine1)
        selectPPLNFileBar.addWidget(
            self._createButton("Select file", self.__openPPLNFile)
        )

        fileLayout = QHBoxLayout(self._parent)

        selectFileBar = QHBoxLayout(self._parent)
        selectFileBar.setSpacing(0)
        selectFileBar.setContentsMargins(0, 0, 5, 0)
        fNameLine = QLineEdit(self._parent)
        fNameLine.setObjectName("csvFile")
        fNameLine.setPlaceholderText("Select a CSV file with features...")
        fNameLine.setReadOnly(True)
        fNameLine.setFont(font)
        selectFileBar.addWidget(fNameLine)
        editBtn = self._createButton(
            None, self._editCSVFile, "editCSVButton", qta.icon("fa5.edit")
        )
        editBtn.setEnabled(False)
        selectFileBar.addWidget(editBtn)
        selectFileBar.addWidget(self._createButton("Select file", self._openCSVFile))

        checkBox = QCheckBox("CSV has header")
        checkBox.setObjectName("csv")
        checkBox.setFont(font)

        fileLayout.addItem(selectFileBar)
        fileLayout.addWidget(checkBox)

        confirmBar = QHBoxLayout(self._parent)
        confirmBar.addStretch()
        confirmBar.addWidget(self._createButton("Run", self.__runPipeline))

        vBoxLayout.addItem(selectPPLNFileBar)
        vBoxLayout.addItem(fileLayout)
        vBoxLayout.addItem(confirmBar)

        self.setLayout(vBoxLayout)

    def __openPPLNFile(self):
        fname = QFileDialog.getOpenFileName(
            parent=self._parent,
            caption="Select pipeline File",
            filter="PPLN files (*.PPLN)",
        )
        self.findChild(QLineEdit, "pplnFile").setText(fname[0])

    def __runPipeline(self):
        err = ""

        pplnSrc = self.findChild(QLineEdit, "pplnFile").text()
        if self._isNoneOrWhiteSpace(pplnSrc):
            err += "Select pipeline file.\n"

        csvSrc = self.findChild(QLineEdit, "csvFile").text()
        if self._isNoneOrWhiteSpace(csvSrc):
            err += "Select CSV file with features.\n"

        if not self._isNoneOrWhiteSpace(err):
            self._parent.errorMessage.setText(err)
            self._parent.errorMessage.show()
            return
        
        self._processWindow = ProcessWindow(
            self._parent,
            ProcessWindowData(
                False,
                csvSrc=csvSrc,
                csvHasHeader=self.findChild(QCheckBox, "csv").isChecked(),
                pipelineSrc=pplnSrc,
            ),
        )
        self._processWindow.show()



import sys, pathlib, os
PROJECT_ROOT = pathlib.Path(__file__).resolve().parent.parent.parent
if str(PROJECT_ROOT) not in sys.path:
    sys.path.insert(0, str(PROJECT_ROOT))
from niaaml_gui.utils.pipeline_runner import run_pipeline

if __name__ == "__main__":
    import json

    args = json.loads(sys.stdin.read())

    def log_fn(msg):
        print(msg, flush=True)
        
    run_pipeline(
        csv_path=args["csv_path"],
        has_header=args["has_header"],
        contains_classes=args["contains_classes"],
        fitness_name=args["fitness_name"],
        pop_size=args["pop_size"],
        inner_pop=args["inner_pop"],
        evals=args["evals"],
        inner_evals=args["inner_evals"],
        opt_alg=args["opt_alg"],
        classifiers=args["classifiers"],
        fs_algorithms=args["fs_algorithms"],
        ft_algorithms=args["ft_algorithms"],
        log_fn=log_fn,
        save_path=pathlib.Path(args["save_path"]),
    )



from PyQt6.QtCore import QThread, pyqtSignal
from niaaml.data import CSVDataReader
from niaaml import PipelineOptimizer
import os

def print_pipeline_values(data):
        print("\n Vrednosti, ki se pošiljajo v optimizacijo:\n")
        for key, value in vars(data).items():
                print(f"  {key} = {value}")
        print("\n" + "-" * 60 + "\n")
class OptimizeThread(QThread):
    optimized = pyqtSignal(object)
    progress = pyqtSignal(object)

    def __init__(self, data):
        super().__init__()
        self.__data = data
    
    def run(self):
        dataReader = CSVDataReader(
            src=self.__data.csvSrc,
            has_header=self.__data.csvHasHeader
        )

        fsas = self.__data.fsas.split("\n") if isinstance(self.__data.fsas, str) else self.__data.fsas
        ftas = self.__data.ftas.split("\n") if isinstance(self.__data.ftas, str) else self.__data.ftas
        clfs = self.__data.classifiers.split("\n") if isinstance(self.__data.classifiers, str) else self.__data.classifiers
        optimizer = PipelineOptimizer(
            data=dataReader,
            classifiers=[c.strip() for c in clfs if c.strip()],
            feature_selection_algorithms=[f.strip() for f in fsas if f.strip()],
            feature_transform_algorithms=[f.strip() for f in ftas if f.strip()],
            categorical_features_encoder=self.__data.encoder,
            imputer=self.__data.imputer,         
        )
        optimizer._PipelineOptimizer__logger = HackyLogger(self.progress.emit)
        print_pipeline_values(self.__data)
        if self.__data.isOptimization is True:
            pipeline = optimizer.run(
                self.__data.fitnessFunctionName,
                self.__data.popSize,
                self.__data.popSizeInner,
                self.__data.numEvals,
                self.__data.numEvalsInner,
                self.__data.optAlgName,
                self.__data.optAlgInnerName,
            )
        else:
            pipeline = optimizer.run_v1(
                self.__data.fitnessFunctionName,
                self.__data.popSize,
                self.__data.numEvals,
                self.__data.optAlgName,
            )
        
        pipeline.export(os.path.join(self.__data.outputFolder, "niaamlGUIoutput"))
        pipeline.export_text(os.path.join(self.__data.outputFolder, "niaamlGUIoutput"))
        
        self.optimized.emit(pipeline.to_string())


class HackyLogger:
    def __init__(self, emit_func, **kwargs):
        self.__emit = emit_func

    def log_progress(self, text):
        self.__emit(text)

    def log_pipeline(self, text):
        return

    def log_optimization_error(self, text):
        return
    
    def info(self, msg):
        print("msg", msg)  



import subprocess
import json
from PyQt6.QtCore import QThread, pyqtSignal
import logging

class PipelineRunnerThread(QThread):
    ran = pyqtSignal(object)
    progress = pyqtSignal(str)

    def __init__(self, data):
        super().__init__()
        self._data = data
        nia_logger = logging.getLogger('NiaAML')
        nia_logger.setLevel(logging.INFO)
        
        class _QtHandler(logging.Handler):
            def emit(inner_self, record):
                self.progress.emit(record.getMessage())

        if not any(isinstance(h, _QtHandler) for h in nia_logger.handlers):
            nia_logger.addHandler(_QtHandler())

    def run(self):
        try:
            args = {
                "csv_path": self._data.csvSrc,
                "has_header": self._data.csvHasHeader,
                "contains_classes": True,
                "fitness_name": self._data.fitnessFunctionName,
                "pop_size": int(self._data.popSize or 0),
                "inner_pop": int(self._data.popSizeInner or 0),
                "evals": int(self._data.numEvals or 0),
                "inner_evals": int(self._data.numEvalsInner or 0),
                "opt_alg": self._data.optAlgName,
                "classifiers": self._to_list(self._data.classifiers),
                "fs_algorithms": self._to_list(self._data.fsas),
                "ft_algorithms": self._to_list(self._data.ftas),
                "save_path": self._data.outputFolder
            }

            process = subprocess.Popen(
                ["python", "-m", "niaaml_gui.utils.run_pipeline_subprocess"],
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.STDOUT,
                text=True
            )

            json.dump(args, process.stdin)
            process.stdin.close()

            for line in process.stdout:
                line = line.strip()
                self.progress.emit(line)

            process.wait()

            if process.returncode == 0:
                self.ran.emit("Pipeline finished successfully.")
            else:
                self.ran.emit("ERROR: Subprocess failed.")

        except Exception as e:
            import traceback
            traceback.print_exc()
            self.ran.emit(f"ERROR: {str(e)}")

    def _to_list(self, val):
        if isinstance(val, list):
            return val
        return [v.strip() for v in val.splitlines() if v.strip()]



from PyQt6.QtWidgets import QDialog, QVBoxLayout, QListWidget, QListWidgetItem, QPushButton

class MultiSelectDialog(QDialog):
    def __init__(self, title, options):
        super().__init__()
        self.setWindowTitle(title)
        self.setMinimumWidth(300)

        layout = QVBoxLayout(self)
        self.listWidget = QListWidget()
        self.listWidget.setSelectionMode(QListWidget.SelectionMode.MultiSelection)

        for option in options:
            item = QListWidgetItem(option)
            self.listWidget.addItem(item)

        layout.addWidget(self.listWidget)

        btn = QPushButton("OK")
        btn.clicked.connect(self.accept)
        layout.addWidget(btn)

    def selected_items(self):
        return [item.text() for item in self.listWidget.selectedItems()]



import os

from PyQt6.QtWidgets import (
    QGraphicsView,
    QGraphicsScene,
    QGraphicsRectItem,
    QGraphicsTextItem,
    QGraphicsItem,
    QFileDialog,
    QLineEdit,
    QGraphicsProxyWidget,
    QComboBox,
    QGraphicsPixmapItem,
    QCheckBox,
    QPushButton,
    QMessageBox,
    QGraphicsEllipseItem,
    QGraphicsPathItem
)

from PyQt6.QtGui import (
    QPainter,
    QColor,
    QBrush,
    QPen,
    QIntValidator,
    QFontMetricsF,
    QPixmap,
    QIcon,
    QPainterPath
)

from PyQt6.QtCore import Qt, QSize, QPointF, QRectF, pyqtSignal

from niaaml.classifiers import ClassifierFactory
from niaaml.preprocessing.feature_transform import FeatureTransformAlgorithmFactory
from niaaml.preprocessing.feature_selection import FeatureSelectionAlgorithmFactory
from niaaml.preprocessing.encoding import EncoderFactory
from niaaml.preprocessing.imputation import ImputerFactory
from niaaml.fitness import FitnessFactory

from niapy.util.factory import _algorithm_options

from niaaml_gui.widgets.connection_line import ConnectionLine
from niaaml_gui.widgets.multi_selection_dialog import MultiSelectDialog
from niaaml_gui.windows.csv_editor_window import CSVEditorWindow

class PipelineCanvas(QGraphicsView):
    __niaamlFitnessFunctions = FitnessFactory().get_name_to_classname_mapping()
    __niaamlEncoders = EncoderFactory().get_name_to_classname_mapping()
    __niaamlImputers = ImputerFactory().get_name_to_classname_mapping()
    __niapyAlgorithmsList = list(_algorithm_options().keys())
    __niaamlFitnessFunctionsList = list(__niaamlFitnessFunctions.keys())
    __niaamlEncodersList = list(__niaamlEncoders.keys())
    __niaamlImputersList = list(__niaamlImputers.keys())
    
    pipelineStateChanged = pyqtSignal()
    
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setRenderHint(QPainter.RenderHint.Antialiasing)
        self.setAcceptDrops(True)
        self.scene = QGraphicsScene(self)
        self.setScene(self.scene)
        self.setAlignment(Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)
        self.setAcceptDrops(True)
        self.block_data = {}
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)
        self.current_connection_line = None
        self.pending_output_block = None
        self.line_start = None
        self._highlighted_circles = []
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        self.setResizeAnchor(QGraphicsView.ViewportAnchor.NoAnchor)
        self.setSceneRect(0, 0, 5000, 5000) 
        
    def get_value(self):
        return self.value

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dragMoveEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event):
        if event.mimeData().hasText():
            text = event.mimeData().text()
            pos = self.mapToScene(int(event.position().x()), int(event.position().y()))
            self.add_config_block(text, pos.x(), pos.y())
            event.acceptProposedAction()

    def add_config_block(self, label: str, x: int = 50, y: int = 50):
        categories = {
            "Data": {
                "Select CSV File", "Pipeline Output Folder"
            },
            "Pre-processing": {
                "Categorical Encoder", "Missing Imputer"
            },
            "Feature Eng.": {
                "Feature Selection", "Feature Transform"
            },
            "Modeling": {
                "Classifier",
                "Optimization Algorithm (Selection)",
                "Optimization Algorithm (Tuning)"
            },
            "Population": {
                "Population Size (Components Selection)", "Population Size (Parameter Tuning)"
            },
            "Evaluation": {
                "Number of Evaluations (Component Selection)", "Number of Evaluations (Parameter Tuning)"
            },
            "Fitness": {
                "Fitness Function"
            }
        }
        def _category_for(lbl: str) -> str:
            for cat, items in categories.items():
                if lbl in items:
                    return cat
            return "Data"
        cat = _category_for(label)

        category_shapes = {
            "Data": "rounded",                      
            "Pre-processing": "cut-corner",        
            "Feature Eng.": "parallelogram",             
            "Modeling": "hexagon",                
            "Population": "octagon",               
            "Evaluation": "rounded",             
            "Fitness": "ellipse"                   
        }

        shape = category_shapes.get(cat, "rect")

        is_file = label == "Select CSV File"
        is_folder = label == "Pipeline Output Folder"
        is_number_input = label in [
            "Population Size (Components Selection)",
            "Population Size (Parameter Tuning)",
            "Number of Evaluations (Component Selection)",
            "Number of Evaluations (Parameter Tuning)",
        ]

        # dropdowni
        dropdown_options = []
        if label == "Missing Imputer":
            dropdown_options = self.__niaamlImputersList
        elif label == "Categorical Encoder":
            dropdown_options = self.__niaamlEncodersList
        elif label in [
            "Optimization Algorithm (Selection)",
            "Optimization Algorithm (Tuning)",
        ]:
            dropdown_options = self.__niapyAlgorithmsList
        elif label == "Fitness Function":
            dropdown_options = self.__niaamlFitnessFunctionsList

        # ikona
        icon_dir = os.path.join(
            os.path.dirname(os.path.dirname(__file__)), "resources", "icons"
        )
        icon_map = {
            "Select CSV File": "file.png",
            "Categorical Encoder": "encode.png",
            "Missing Imputer": "imputation.png",
            "Feature Selection": "feature_selection.png",
            "Feature Transform": "feature_transformation.png",
            "Classifier": "classifier.png",
            "Optimization Algorithm": "optimization_algorithm.png",
            "Fitness Function": "fitness.png",
            "Pipeline Output Folder": "folder.png",
            "Optimization Algorithm (Selection)": "optimization_algorithm.png",
            "Optimization Algorithm (Tuning)": "decision_tree.png",
            "Population Size (Components Selection)": "numbers.png",
            "Population Size (Parameter Tuning)": "numbers.png",
            "Number of Evaluations (Component Selection)": "numbers.png",
            "Number of Evaluations (Parameter Tuning)": "numbers.png",
        }
        icon_path = os.path.join(icon_dir, icon_map.get(label, "placeholder.png"))

        if is_number_input:
            block = NumericInputBlock(label, icon_path=icon_path, shape=shape)
        elif dropdown_options:
            block = InteractiveConfigBlock(
                label, dropdown_options=dropdown_options, icon_path=icon_path, shape=shape
            )
        else:
            block = InteractiveConfigBlock(
                label, is_file=is_file, is_folder=is_folder,
                icon_path=icon_path, shape=shape
            )

        block.setPos(x, y)
        self.scene.addItem(block)
        self.block_data[block] = {"label": label, "path": None}
        self.pipelineStateChanged.emit()
        
        if hasattr(block, "input_field"):
            block.input_field.textChanged.connect(self.pipelineStateChanged.emit)
        elif hasattr(block, "dropdown"):
            block.dropdown.currentTextChanged.connect(self.pipelineStateChanged.emit)

    def handle_block_click(self, block):
        label = block.label

        if "file" in label.lower():
            path, _ = QFileDialog.getOpenFileName(self, "Select File")
            if path:
                self.block_data[block] = {"label": label, "path": path}
                block.path_display.setPlainText(os.path.basename(path))
            self.pipelineStateChanged.emit()

        elif "folder" in label.lower():
            path = QFileDialog.getExistingDirectory(self, "Select Folder")
            if path:
                self.block_data[block] = {"label": label, "path": path}
                block.path_display.setPlainText(os.path.basename(path))
            self.pipelineStateChanged.emit()

        else:
            pass

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Delete:
            for item in self.scene.selectedItems():
                if isinstance(item, ConnectionLine):
                    if hasattr(item.source_block, "connections"):
                        if item in item.source_block.connections:
                            item.source_block.connections.remove(item)
                    if hasattr(item.target_block, "connections"):
                        if item in item.target_block.connections:
                            item.target_block.connections.remove(item)
                self.scene.removeItem(item)
            self.pipelineStateChanged.emit()
            
        else:
            super().keyPressEvent(event)


    def progress_start(self, maximum: int | None = None) -> None:
        if maximum is None:
            self._progressBar.setMaximum(0)
            self._progressBar.setTextVisible(False)
        else:
            self._progressBar.setMaximum(maximum)
            self._progressBar.setValue(0)
            self._progressBar.setTextVisible(True)

        r = self.sceneRect()
        self._progressBar.setGeometry(
            int(r.x()), int(r.bottom() - 22), int(r.width()), 18
        )
        self._progressBar.show()
        self.update()

    def progress_set(self, value: int) -> None:
        if self._progressBar.isVisible() and self._progressBar.maximum() != 0:
            self._progressBar.setValue(value)

    def progress_finish(self) -> None:
        self._progressBar.hide()
        
    def mousePressEvent(self, event):
        view_pos  = event.pos()               
        scene_pos = self.mapToScene(view_pos) 
        item = self.itemAt(view_pos)           

        if isinstance(item, QGraphicsEllipseItem):
            item = item.parentItem()

        if isinstance(item, (InteractiveConfigBlock, NumericInputBlock)):
            local = item.mapFromScene(scene_pos)
            if item.output_circle.contains(local):
                self.pending_output_block = item

                self.line_preview = QGraphicsPathItem()
                self.line_preview.setZValue(99)
                self.line_preview.setPen(QPen(QColor("black"), 2, Qt.PenStyle.DashLine))
                self.line_preview.setFlag(QGraphicsPathItem.GraphicsItemFlag.ItemIgnoresTransformations)
                self.line_preview.setFlag(QGraphicsPathItem.GraphicsItemFlag.ItemDoesntPropagateOpacityToChildren)

                self.scene.addItem(self.line_preview)
                self.line_start = item.output_point()

                self._highlight_connection_targets(item)
                return  

        super().mousePressEvent(event)


    def mouseMoveEvent(self, event):
        if getattr(self, "line_preview", None):
            end = self.mapToScene(event.pos())
            path = QPainterPath()
            path.moveTo(self.line_start)
            dx = (end.x() - self.line_start.x()) * 0.5
            cp1 = QPointF(self.line_start.x() + dx, self.line_start.y())
            cp2 = QPointF(end.x() - dx, end.y())
            path.cubicTo(cp1, cp2, end)
            self.line_preview.setPath(path)
        else:
            super().mouseMoveEvent(event)


    def mouseReleaseEvent(self, event):
        if getattr(self, "line_preview", None):
            self.scene.removeItem(self.line_preview)
            self.line_preview = None

        self._clear_connection_highlights()

        if self.pending_output_block:
            view_pos = event.pos()
            scene_pos = self.mapToScene(view_pos)
            item = self.scene.itemAt(scene_pos, self.transform())
            if isinstance(item, QGraphicsEllipseItem):
                item = item.parentItem()

            if isinstance(item, (InteractiveConfigBlock, NumericInputBlock))  and item is not self.pending_output_block:
                local = item.mapFromScene(scene_pos)
                if item.input_circle.contains(local):
                    if self._is_valid_connection(self.pending_output_block, item):
                        connection = ConnectionLine(self.pending_output_block, item)
                        self.scene.addItem(connection)
                        self.pending_output_block.add_connection(connection)
                        item.add_connection(connection)

            self.pending_output_block = None
            self.pipelineStateChanged.emit()

        else:
            super().mouseReleaseEvent(event)
            self.pipelineStateChanged.emit()

            
    def _is_valid_connection(self, source, target) -> bool:
        if source is target:
            return False

        source_label = getattr(source, "label", "").strip().lower()
        target_label = getattr(target, "label", "").strip().lower()

        categories = [
            ("data", {"select csv file", "pipeline output folder"}),
            ("pre-processing", {"categorical encoder", "missing imputer"}),
            ("feature eng.", {"feature selection", "feature transform"}),
            ("modeling", {"classifier", "optimization algorithm (selection)", "optimization algorithm (tuning)"}),
            ("population", {"population size (components selection)", "population size (parameter tuning)"}),
            ("evaluation", {"number of evaluations (component selection)", "number of evaluations (parameter tuning)"}),
            ("fitness", {"fitness function"})
        ]

        def find_category(label: str) -> str | None:
            for cat, items in categories:
                if label in items:
                    return cat
            return None

        source_cat = find_category(source_label)
        target_cat = find_category(target_label)

        if not source_cat or not target_cat:
            return False

        
        if source_label == "pipeline output folder":
            return False
        
        if target_label == "select csv file":
            return False
    
        if source_cat == "fitness" and target_label == "pipeline output folder":
            return True
     
        if source_cat == target_cat:
            return True

        cat_order = [cat for cat, _ in categories]
        try:
            source_idx = cat_order.index(source_cat)
            target_idx = cat_order.index(target_cat)
            if target_idx == source_idx + 1:
                return True
        except ValueError:
            return False

        return False


    
    def _highlight_connection_targets(self, source_block):
        for item in self.scene.items():
            if item is source_block:
                continue

            if isinstance(item, (InteractiveConfigBlock, NumericInputBlock)):
                is_valid = self._is_valid_connection(source_block, item)
                color = QColor("green") if is_valid else QColor("red")
                item.input_circle.setBrush(QBrush(color))
                self._highlighted_circles.append(item.input_circle)


    def _clear_connection_highlights(self):
        for circle in self._highlighted_circles:
            circle.setBrush(QBrush(QColor("white")))
        self._highlighted_circles.clear()        
        
    def is_pipeline_ready(self) -> bool:
        if not self.block_data:
            return False

        for block, info in self.block_data.items():
            if hasattr(block, "get_value"):
                value = block.get_value()
                if not value or (isinstance(value, str) and not value.strip()):
                    return False
            elif info.get("path") is None:
                return False
        return True




class InteractiveConfigBlock(QGraphicsPathItem):
    __niaamlFeatureSelectionAlgorithmsMap = (
        FeatureSelectionAlgorithmFactory().get_name_to_classname_mapping()
    )
    __niaamlFeatureTransformAlgorithmsMap = (
        FeatureTransformAlgorithmFactory().get_name_to_classname_mapping()
    )
    __niaamlClassifiersMap = ClassifierFactory().get_name_to_classname_mapping()

    __niaamlFeatureSelectionDisplayOptions = list(__niaamlFeatureSelectionAlgorithmsMap)
    __niaamlFeatureTransformDisplayOptions = list(__niaamlFeatureTransformAlgorithmsMap)
    __niaamlClassifiersDisplayOptions      = list(__niaamlClassifiersMap)

    def __init__(
        self,
        label: str,
        *,
        shape: str = "rect",
        value: str = "",
        is_file: bool = False,
        is_folder: bool = False,
        is_number_input: bool = False,
        dropdown_options=None,
        icon_path: str | None = None,
        is_sidebar=True,
        readonly=False  
    ):
        super().__init__()
        self.readonly = readonly  
        self.shape            = shape.lower()
        self.icon_path        = icon_path
        self.label            = label
        self.value            = value
        self.is_file          = is_file
        self.is_folder        = is_folder
        self.is_number_input  = is_number_input
        self.dropdown_options = dropdown_options or []
        self.selected_options : list[str] = []
        self.connections      : list      = []
        
        tmp = QComboBox(); tmp.addItems(self.dropdown_options)
        combo_w   = tmp.sizeHint().width() if self.dropdown_options else 0
        self.w    = max(220, combo_w + 20)
        self.h    = 80

        self._set_shape_path()
        self.setBrush(QBrush(QColor("#005f85")))
        self.setPen  (QPen  (QColor("white")))
        self.setFlags(
            QGraphicsPathItem.GraphicsItemFlag.ItemIsMovable            |
            QGraphicsPathItem.GraphicsItemFlag.ItemIsSelectable         |
            QGraphicsPathItem.GraphicsItemFlag.ItemSendsGeometryChanges
        )

        lbl_x = 10
        if self.icon_path and os.path.exists(self.icon_path):
            pm = QPixmap(self.icon_path).scaled(
                20, 20, Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation
            )
            self.icon_item = QGraphicsPixmapItem(pm, self)
            self.icon_item.setOffset(5, 5)
            lbl_x = 30

        self.label_item = QGraphicsTextItem(self)
        self.label_item.setDefaultTextColor(QColor("white"))
        self.label_item.setPlainText(self.label)
        self.label_item.setPos(lbl_x, 5)

        y = 30
        if self.dropdown_options:
            dd = QComboBox(); dd.addItems(self.dropdown_options)
            pr = QGraphicsProxyWidget(self); pr.setWidget(dd); pr.setPos(10, y)
            self.dropdown = dd
            y += dd.sizeHint().height() + 10

        elif self.is_number_input:
            le = QLineEdit(); le.setValidator(QIntValidator(0, 999999)); le.setFixedWidth(100)
            pr = QGraphicsProxyWidget(self); pr.setWidget(le); pr.setPos(10, y)
            self.input_field = le
            y += le.sizeHint().height() + 10

        else:
            self.value_item = QGraphicsTextItem("Click to select…", self)
            self.value_item.setDefaultTextColor(QColor("white"))
            self.value_item.setTextWidth(self.w - 20)  
            self.value_item.setZValue(1)


            text_rect = self.value_item.boundingRect()
            text_x = 10
            text_y = y
            self.value_item.setPos(text_x, text_y)

            self.click_button = QPushButton("")
            self.click_button.setFlat(True)
            self.click_button.setStyleSheet("background-color: transparent; border: none;")
            self.click_button.setCursor(Qt.CursorShape.PointingHandCursor)
            self.click_button.setFixedSize(int(text_rect.width()), int(text_rect.height()))

            self.click_button.clicked.connect(lambda: self._handle_click_action())

            self.click_proxy = QGraphicsProxyWidget(self)
            self.click_proxy.setWidget(self.click_button)
            self.click_proxy.setPos(text_x, text_y)

            if self.is_file and self.label == "Select CSV File":
                self._add_csv_helpers()

        self.input_circle  = QGraphicsEllipseItem(self)
        self.output_circle = QGraphicsEllipseItem(self)
        for circ in (self.input_circle, self.output_circle):
            circ.setBrush(QBrush(QColor("white")))
            circ.setZValue(1)
        self._reposition_io()
        self._layout_contents()

    def _handle_click_action(self):
        
        if self.readonly:
            return
        if self.dropdown_options:
            return 
        elif self.label in ["Feature Selection", "Feature Transform", "Classifier"]:
            self.getMultiSelection()
        elif self.is_file or self.is_folder:
            self.getPath()
        
    def _set_shape_path(self):
        """Definira QPainterPath glede na izbrano obliko."""
        p = QPainterPath()
        w, h = self.w, self.h

        if self.shape == "rounded":
            p.addRoundedRect(0, 0, w, h, 10, 10)

        elif self.shape == "diamond":
            p.moveTo(w / 2, 0)
            p.lineTo(w, h / 2)
            p.lineTo(w / 2, h)
            p.lineTo(0, h / 2)
            p.closeSubpath()

        elif self.shape == "octagon":
            r = 10
            p.moveTo(r, 0)
            p.lineTo(w - r, 0)
            p.lineTo(w, r)
            p.lineTo(w, h - r)
            p.lineTo(w - r, h)
            p.lineTo(r, h)
            p.lineTo(0, h - r)
            p.lineTo(0, r)
            p.closeSubpath()

        elif self.shape == "hexagon":
            r = 15
            p.moveTo(r, 0)
            p.lineTo(w - r, 0)
            p.lineTo(w, h / 2)
            p.lineTo(w - r, h)
            p.lineTo(r, h)
            p.lineTo(0, h / 2)
            p.closeSubpath()

        elif self.shape == "cut-corner":
            r = 10
            p.moveTo(r, 0)
            p.lineTo(w - r, 0)
            p.lineTo(w, r)
            p.lineTo(w, h - r)
            p.lineTo(w - r, h)
            p.lineTo(r, h)
            p.lineTo(0, h - r)
            p.lineTo(0, r)
            p.closeSubpath()

        elif self.shape == "notched":
            notch = 10
            p.moveTo(notch, 0)
            p.lineTo(w - notch, 0)
            p.lineTo(w, notch)
            p.lineTo(w, h - notch)
            p.lineTo(w - notch, h)
            p.lineTo(notch, h)
            p.lineTo(0, h - notch)
            p.lineTo(0, notch)
            p.closeSubpath()
        elif self.shape == "notched":
            notch = 10
            p.moveTo(notch, 0)
            p.lineTo(w - notch, 0)
            p.lineTo(w, notch)
            p.lineTo(w, h - notch)
            p.lineTo(w - notch, h)
            p.lineTo(notch, h)
            p.lineTo(0, h - notch)
            p.lineTo(0, notch)
            p.closeSubpath()
        elif self.shape == "parallelogram":
            slant = 20
            p.moveTo(0, 0)
            p.lineTo(self.w - slant, 0)
            p.lineTo(self.w, self.h)
            p.lineTo(slant, self.h)
            p.closeSubpath()
        else:  
            p.addRect(0, 0, w, h)

        self.setPath(p)

    def _reposition_io(self):
        self.input_circle .setRect(-6,        self.h/2 - 6, 12, 12)
        self.output_circle.setRect(self.w - 6, self.h/2 - 6, 12, 12)

    def _add_csv_helpers(self):
        self.csvHasHeader = True
        cb = QCheckBox("Header"); cb.setChecked(True)
        p_cb = QGraphicsProxyWidget(self); p_cb.setWidget(cb)
        p_cb.setPos(self.w - cb.sizeHint().width() - 5, 5)
        cb.stateChanged.connect(
            lambda s: setattr(self, "csvHasHeader", s == Qt.CheckState.Checked)
        )
        
        icon_dir  = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "icons")
        edit_icon = os.path.join(icon_dir, "edit.png")
        btn = QPushButton(); btn.setIcon(QIcon(edit_icon)); btn.setIconSize(QSize(18, 18))
        btn.setFixedSize(24, 24)
        p_btn = QGraphicsProxyWidget(self); p_btn.setWidget(btn)
        p_btn.setPos(self.w - 24 - 5, self.h - 24 - 5)
        btn.clicked.connect(self._open_csv_editor)
        self._edit_btn_proxy = p_btn

    def input_point (self) -> QPointF: return self.mapToScene(0,      self.h/2)
    def output_point(self) -> QPointF: return self.mapToScene(self.w, self.h/2)

    def _open_csv_editor(self):
        if not getattr(self, "value", ""):
            QMessageBox.warning(None, "No CSV", "Please pick a CSV file first.")
            return
        self.csv_editor_window = CSVEditorWindow(self.value)
        self.csv_editor_window.show()

    def getPath(self):
        dlg = QFileDialog
        if self.is_file:
            path, _ = dlg.getOpenFileName(None, "Select File", "", "CSV Files (*.csv);;All Files (*)")
        else:
            path = dlg.getExistingDirectory(None, "Select Folder")
        if path:
            self.value = path
            self.update_value_display()

    def getMultiSelection(self):
        if self.label == "Feature Selection":
            disp, mp = self.__niaamlFeatureSelectionDisplayOptions, self.__niaamlFeatureSelectionAlgorithmsMap
        elif self.label == "Feature Transform":
            disp, mp = self.__niaamlFeatureTransformDisplayOptions, self.__niaamlFeatureTransformAlgorithmsMap
        elif self.label == "Classifier":
            disp, mp = self.__niaamlClassifiersDisplayOptions     , self.__niaamlClassifiersMap
        else:
            disp, mp = [], {}

        dlg = MultiSelectDialog(f"Select {self.label}(s)", disp)
        if dlg.exec():
            chosen = dlg.selected_items()
            self.selected_options = [mp[d] for d in chosen if d in mp]
            txt = "\n".join(chosen)
            self.value_item.setPlainText(txt)
            self.value = "\n".join(self.selected_options)
            self._adjust_height(txt)

    def update_value_display(self):
        if hasattr(self, "value_item"):
            if self.selected_options: txt = "\n".join(self.selected_options)
            elif self.value:          txt = self.value
            else:                     txt = "Click to select…"
            self.value_item.setPlainText(txt)
            self._adjust_height(txt)
            
        if hasattr(self, "click_button"):
            text_rect = self.value_item.boundingRect()
            text_x = (self.w - text_rect.width()) / 2
            text_y = self.value_item.pos().y()
            self.click_button.setFixedSize(int(text_rect.width()), int(text_rect.height()))
            if hasattr(self, "click_proxy"):
                self.click_proxy.setPos(text_x, text_y)

    def _adjust_height(self, text: str):
        new_h = self.calculate_block_height(text)
        if new_h != self.h:
            dh = new_h - self.h
            self.h = new_h
            self._set_shape_path()
            self._reposition_io()
            if hasattr(self, "click_proxy"):
                p = self.click_proxy.pos()
                self.click_proxy.setPos(p.x(), p.y() + dh)


    def itemChange(self, ch, val):
        if ch == QGraphicsPathItem.GraphicsItemChange.ItemPositionChange:
            for c in self.connections:
                c.update_path()
        return super().itemChange(ch, val)

    @staticmethod
    def calculate_block_height(text: str, width: int = 200, base: int = 70) -> int:
        fm = QFontMetricsF(QGraphicsTextItem().font())
        line_h = fm.lineSpacing()
        lines  = text.split("\n")
        tot = sum(max(1, int(fm.horizontalAdvance(ln) / width) + 1) for ln in lines)
        return max(base, int(30 + tot * line_h + 20))

    def add_connection(self, conn):
        if conn not in self.connections:
            self.connections.append(conn)
    
    def _layout_contents(self):
        margin = 10
        icon_size = 20
        spacing = 5
        label_x = margin

        if hasattr(self, "icon_item"):
            self.icon_item.setOffset(label_x, margin)
            label_x += icon_size + spacing

        text_width = self.label_item.boundingRect().width()
        self.label_item.setPos((self.w - text_width) / 2, margin)

        if hasattr(self, "input_field"):
            proxy = QGraphicsProxyWidget(self)
            proxy.setWidget(self.input_field)
            field_y = margin + self.label_item.boundingRect().height() + spacing
            proxy.setPos((self.w - self.input_field.width()) / 2, field_y)

    def boundingRect(self):
        return QRectF(0, 0, self.w, self.h)
            


class NumericInputBlock(QGraphicsPathItem):
    def __init__(
        self,
        label: str,
        *,
        shape: str = "rect",
        icon_path: str | None = None
    ):
        super().__init__()
        self.label = label
        self.icon_path = icon_path
        self.shape = shape.lower()
        self.connections = []

        self._w = 220
        self._h = 70

        label_x = 10
        if self.icon_path and os.path.exists(self.icon_path):
            pm = QPixmap(self.icon_path).scaled(
                20, 20,
                Qt.AspectRatioMode.KeepAspectRatio,
                Qt.TransformationMode.SmoothTransformation
            )
            self.icon_item = QGraphicsPixmapItem(pm, self)
            self.icon_item.setOffset(5, 5)
            label_x = 30

        self.label_item = QGraphicsTextItem(self)
        self.label_item.setDefaultTextColor(QColor("white"))
        self.label_item.setTextWidth(self._w - label_x - 10)
        self.label_item.setPlainText(self.label)

        edit_y = 10 + self.label_item.boundingRect().height()
        self.input_field = QLineEdit()
        self.input_field.setValidator(QIntValidator(0, 999999))
        self.input_field.setFixedWidth(100)

        proxy = QGraphicsProxyWidget(self)
        proxy.setWidget(self.input_field)
        proxy.setPos(10, edit_y)

        self._h = edit_y + self.input_field.sizeHint().height() + 10

        self._set_shape_path()

        self.input_circle = QGraphicsEllipseItem(-6, self._h / 2 - 6, 12, 12, self)
        self.output_circle = QGraphicsEllipseItem(self._w - 6, self._h / 2 - 6, 12, 12, self)
        for circ in (self.input_circle, self.output_circle):
            circ.setBrush(QBrush(QColor("white")))
            circ.setZValue(1)

        self.setBrush(QBrush(QColor("#005f85")))
        self.setPen(QPen(QColor("white")))
        self.setFlag(QGraphicsPathItem.GraphicsItemFlag.ItemIsMovable)
        self.setFlag(QGraphicsPathItem.GraphicsItemFlag.ItemIsSelectable)
        self.setFlag(QGraphicsPathItem.GraphicsItemFlag.ItemSendsGeometryChanges)

        self._layout_contents()

    def _set_shape_path(self):
        p = QPainterPath()
        if self.shape == "rounded":
            p.addRoundedRect(0, 0, self._w, self._h, 12, 12)
        elif self.shape == "ellipse":
            p.addEllipse(0, 0, self._w, self._h)
        elif self.shape == "diamond":
            p.moveTo(self._w / 2, 0)
            p.lineTo(self._w, self._h / 2)
            p.lineTo(self._w / 2, self._h)
            p.lineTo(0, self._h / 2)
            p.closeSubpath()
        elif self.shape == "octagon":
            r = 12
            p.moveTo(r, 0)
            p.lineTo(self._w - r, 0)
            p.lineTo(self._w, r)
            p.lineTo(self._w, self._h - r)
            p.lineTo(self._w - r, self._h)
            p.lineTo(r, self._h)
            p.lineTo(0, self._h - r)
            p.lineTo(0, r)
            p.closeSubpath()
        elif self.shape == "hexagon":
            p.moveTo(self._w * 0.25, 0)
            p.lineTo(self._w * 0.75, 0)
            p.lineTo(self._w, self._h * 0.5)
            p.lineTo(self._w * 0.75, self._h)
            p.lineTo(self._w * 0.25, self._h)
            p.lineTo(0, self._h * 0.5)
            p.closeSubpath()
        elif self.shape == "trapezoid":
            p.moveTo(self._w * 0.2, 0)
            p.lineTo(self._w * 0.8, 0)
            p.lineTo(self._w, self._h)
            p.lineTo(0, self._h)
            p.closeSubpath()
        else:
            p.addRect(0, 0, self._w, self._h)
        self.setPath(p)

    def input_point(self) -> QPointF:
        return self.mapToScene(self.input_circle.boundingRect().center())

    def output_point(self) -> QPointF:
        return self.mapToScene(self.output_circle.boundingRect().center())

    def add_connection(self, conn):
        if conn not in self.connections:
            self.connections.append(conn)

    def itemChange(self, change, value):
        if change == QGraphicsPathItem.GraphicsItemChange.ItemPositionChange:
            for conn in self.connections:
                conn.update_path()
        return super().itemChange(change, value)

    def get_value(self) -> str:
        return self.input_field.text()

    def _layout_contents(self):
        margin = 10
        icon_size = 20
        spacing = 5
        label_x = margin

        if hasattr(self, "icon_item"):
            self.icon_item.setOffset(label_x, margin)
            label_x += icon_size + spacing

        if hasattr(self, "label_item"):
            if hasattr(self, "icon_item"):
                self.label_item.setPos(label_x, margin)
            else:
                text_width = self.label_item.boundingRect().width()
                self.label_item.setPos((self._w - text_width) / 2, margin)


    def boundingRect(self):
        return QRectF(0, 0, self._w, self._h)



from PyQt6.QtCore import QThread, pyqtSignal
from niaaml.data import CSVDataReader
from niaaml import Pipeline
from pathlib import Path


class RunThread(QThread):
    ran = pyqtSignal(object)

    def __init__(self, data):
        super().__init__()
        self.__data = data

    def run(self):
        try:
            dataReader = CSVDataReader(
                src=self.__data.csvSrc,
                contains_classes=True,
                has_header=self.__data.csvHasHeader,
            )

            if not self.__data.pipelineSrc:
                self.__data.pipelineSrc = (
                    Path(self.__data.outputFolder) / "niaamlGUIoutput.ppln"
                )

            pipeline = Pipeline.load(self.__data.pipelineSrc)

            x_data = dataReader.get_x()

            predictions = pipeline.run(x_data)

            self.ran.emit(str(predictions))

        except Exception as e:
            import traceback

            traceback.print_exc()
            self.ran.emit(f"ERROR: {str(e)}")



class ProcessWindowData:
    def __init__(
        self,
        isOptimization,
        csvSrc=None,
        csvHasHeader=None,
        encoder=None,
        imputer=None,
        optAlgName=None,
        optAlgInnerName=None,
        popSize=None,
        popSizeInner=None,
        numEvals=None,
        numEvalsInner=None,
        fsas=None,
        ftas=None,
        classifiers=None,
        fitnessFunctionName=None,
        outputFolder=None,
        pipelineSrc=None,
    ):
        self.isOptimization = isOptimization
        self.csvSrc = csvSrc
        self.csvHasHeader = csvHasHeader
        self.encoder = encoder
        self.imputer = imputer
        self.optAlgName = optAlgName
        self.optAlgInnerName = optAlgInnerName
        self.popSize = popSize
        self.popSizeInner = popSizeInner
        self.numEvals = numEvals
        self.numEvalsInner = numEvalsInner
        self.fsas = fsas if fsas is not None and len(fsas) > 0 else None
        self.ftas = ftas if ftas is not None and len(ftas) > 0 else None
        self.classifiers = classifiers
        self.fitnessFunctionName = fitnessFunctionName
        self.outputFolder = outputFolder
        self.pipelineSrc = pipelineSrc
        
    @staticmethod
    def from_dict(d):
        return ProcessWindowData(
            isOptimization=d.get("isOptimization", False),
            csvSrc=d.get("csvSrc"),
            csvHasHeader=d.get("csvHasHeader"),
            encoder=d.get("Categorical Encoder", ""),
            imputer=d.get("Missing Imputer", ""),
            optAlgName=d.get("Optimization Algorithm (Selection)", ""),
            optAlgInnerName=d.get("Optimization Algorithm (Tuning)", ""),
            popSize=d.get("Population Size (Components Selection)", ""),
            popSizeInner=d.get("Population Size (Parameter Tuning)", ""),
            numEvals=d.get("Number of Evaluations (Component Selection)", ""),
            numEvalsInner=d.get("Number of Evaluations (Parameter Tuning)", ""),
            fsas=d.get("Feature Selection", ""),
            ftas=d.get("Feature Transform", ""),
            classifiers=d.get("Classifier", ""),
            fitnessFunctionName=d.get("Fitness Function", ""),
            outputFolder=d.get("Pipeline Output Folder", ""),
            pipelineSrc=d.get("pipelineSrc", "")
        )
        


from niaaml_gui import widgets
from niaaml_gui import windows
from niaaml_gui.main import MainAppWindow
from niaaml_gui.process_window_data import ProcessWindowData
from niaaml_gui.progress_bar import ProgressBar

__all__ = ["widgets", "windows", "ProcessWindowData", "ProgressBar", "MainAppWindow"]
__version__ = "0.4.3"



from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QVBoxLayout, QLabel, QSpacerItem

from niaaml_gui.widgets.base_main_widget import BaseMainWidget


class HelpAuthorsWidget(BaseMainWidget):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        vBoxLayout = QVBoxLayout(self._parent)
        vBoxLayout.setAlignment(Qt.AlignmentFlag.AlignTop)

        maintainers_title = QLabel("Maintainers")
        font = QFont()
        font.setPointSize(20)
        maintainers_title.setFont(font)

        vBoxLayout.addWidget(maintainers_title)

        contentFont = QFont()
        contentFont.setPointSize(12)

        maintainers_content = QLabel()
        maintainers_content.setText("""
            <html>
                <ul>
                    <li>Iztok Fister, Jr. @firefly-cpp</li>
                </ul>
            </html>
        """)
        maintainers_content.setFont(contentFont)

        maintainers_content.setStyleSheet("""
            QLabel {
                padding-bottom: 12px;
            }
        """)

        vBoxLayout.addWidget(maintainers_content)

        contributors_title = QLabel("Contributors (in alphabetical order by surname)")
        font = QFont()
        font.setPointSize(20)
        contributors_title.setFont(font)

        vBoxLayout.addWidget(contributors_title)

        contributors_content = QLabel()
        contributors_content.setText("""
            <html>
                <ul>
                    <li>Ben Beasley @musicinmybrain</li>
                    <li>Zala Lahovnik @zala-lahovnik</li>
                    <li>Luka Pečnik @lukapecnik</li>
                    <li>Žiga Stupan @zStupan</li>
                </ul>
            </html>
        """)
        contributors_content.setFont(contentFont)

        vBoxLayout.addWidget(contributors_content)

        self.setLayout(vBoxLayout)



from PyQt6.QtWidgets import QWidget, QPushButton, QVBoxLayout
from PyQt6.QtCore import pyqtSignal

class PipelineControlsWidget(QWidget):
    runClicked = pyqtSignal()
    resetClicked = pyqtSignal()
    savePipelineClicked = pyqtSignal()
    loadPipelineClicked = pyqtSignal()
    exportPipelineClicked = pyqtSignal()
    validatePipelineClicked = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        layout = QVBoxLayout(self)
        layout.setContentsMargins(10, 10, 10, 10)
        layout.setSpacing(8)

        self.run_button = QPushButton("▶ Run Pipeline")
        self.run_button.setObjectName("run_button")
        self.run_button.clicked.connect(self.runClicked.emit)
        self.setRunEnabled(False)
        layout.addWidget(self.run_button)
        
        self.reset_button = QPushButton("🔄 Reset Pipeline")
        self.reset_button.clicked.connect(self.resetClicked.emit)
        layout.addWidget(self.reset_button)

        self.save_button = QPushButton("💾 Save Pipeline")
        self.save_button.clicked.connect(self.savePipelineClicked.emit)
        layout.addWidget(self.save_button)

        self.load_button = QPushButton("📂 Load Pipeline")
        self.load_button.clicked.connect(self.loadPipelineClicked.emit)
        layout.addWidget(self.load_button)

        self.export_button = QPushButton("📤 Export Results")
        self.export_button.clicked.connect(self.exportPipelineClicked.emit)
        layout.addWidget(self.export_button)

        layout.addStretch()

    def setRunEnabled(self, enabled: bool):
        self.run_button.setEnabled(enabled)
        
        if enabled:
            self.run_button.setStyleSheet("""
                QPushButton {
                    background-color: #008569;
                    color: white;
                    border: 1px solid #006f59;
                    border-radius: 6px;
                    padding: 5px;
                    font-weight: 600;
                }
                QPushButton:hover {
                    background-color: #001D85;
                }
            """)
        else:
            self.run_button.setStyleSheet("""
                QPushButton {
                    background-color: #444;
                    color: #aaa;
                    border: 1px solid #222;
                    border-radius: 6px;
                    padding: 5px;
                    font-weight: 600;
                }
            """)



from PyQt6.QtGui import QStandardItem, QStandardItemModel
from PyQt6.QtWidgets import (
    QMainWindow,
    QWidget,
    QHBoxLayout,
    QTableView,
    QMessageBox,
    QVBoxLayout,
    QPushButton,
)
from PyQt6.QtCore import QSize, Qt
import csv
import qtawesome as qta


class CSVEditorWindow(QMainWindow):
    def __init__(self, src):
        super(CSVEditorWindow, self).__init__()
        self.setMinimumSize(QSize(640, 480))

        self.__src = src
        self.__table = QTableView(self)
        self.__model = QStandardItemModel(self)
        self.__table.setModel(self.__model)

        try:
            with open(src, "r") as f:
                for row in csv.reader(f):
                    items = [QStandardItem(field) for field in row]
                    self.__model.appendRow(items)
        except OSError:
            self.__errorMessage = QMessageBox(
                QMessageBox.Icon.Critical, "Error", "File could not be read."
            )
            self.__errorMessage.show()
            return

        centralWidget = QWidget(self)
        layout = QVBoxLayout(centralWidget)

        toolBar = QHBoxLayout(centralWidget)
        toolBar.setAlignment(Qt.AlignmentFlag.AlignLeft)
        saveBtn = self.__createButton(None, self.__save, None, qta.icon("fa5.save", color_off='white'))
        toolBar.addWidget(saveBtn)

        layout.addItem(toolBar)
        layout.addWidget(self.__table)

        centralWidget.setLayout(layout)
        self.setCentralWidget(centralWidget)

    def __createButton(self, text, callback=None, objectName=None, icon=None):
        btn = QPushButton(self)
        btn.setText(text)
        font = btn.font()
        font.setPointSize(12)
        btn.setFont(font)

        if callback is not None:
            btn.clicked.connect(callback)

        if objectName is not None:
            btn.setObjectName(objectName)

        if icon is not None:
            btn.setIcon(icon)
            btn.setIconSize(QSize(21, 21))

        return btn

    def __save(self):
        self.__writeCsv(self.__src)

    def keyPressEvent(self, e):
        if e.key() == Qt.Key.Key_Delete or e.key() == Qt.Key.Key_Backspace:
            rows = self.__table.selectionModel().selectedRows()
            for index in rows:
                self.__model.removeRow(index.row())

            cols = self.__table.selectionModel().selectedColumns()
            for index in cols:
                self.__model.removeColumn(index.column())

            cells = self.__table.selectedIndexes()
            for index in cells:
                if self.__model.item(index.row(), index.column()) is not None:
                    self.__model.item(index.row(), index.column()).setText(None)

    def __writeCsv(self, fileName):
        try:
            with open(fileName, "w", newline="") as fileOutput:
                writer = csv.writer(fileOutput)
                for rowNumber in range(self.__model.rowCount()):
                    fields = [
                        self.__model.data(
                            self.__model.index(rowNumber, columnNumber),
                            Qt.ItemDataRole.DisplayRole,
                        )
                        for columnNumber in range(self.__model.columnCount())
                    ]
                    writer.writerow(fields)
        except OSError:
            self.__errorMessage = QMessageBox(
                QMessageBox.Icon.Critical, "Error", "File could not be saved."
            )
            self.__errorMessage.show()
            return



from PyQt6.QtWidgets import (
    QComboBox,
    QLineEdit,
    QLabel,
    QHBoxLayout,
    QVBoxLayout,
    QGridLayout,
    QTabWidget,
    QFileDialog,
    QCheckBox,
)
from PyQt6.QtGui import QRegularExpressionValidator
from PyQt6.QtCore import QRegularExpression, Qt
from niapy.util.factory import _algorithm_options
from niaaml_gui.widgets.list_widget_custom import ListWidgetCustom
from niaaml_gui.widgets.base_main_widget import BaseMainWidget
from niaaml_gui.windows import ProcessWindow
from niaaml_gui.process_window_data import ProcessWindowData
from niaaml.classifiers import ClassifierFactory
from niaaml.preprocessing.feature_selection import FeatureSelectionAlgorithmFactory
from niaaml.preprocessing.feature_transform import FeatureTransformAlgorithmFactory
from niaaml.fitness import FitnessFactory
from niaaml.preprocessing.encoding import EncoderFactory
from niaaml.preprocessing.imputation import ImputerFactory
import qtawesome as qta


class OptimizationWidget(BaseMainWidget):
    __niaamlFeatureSelectionAlgorithms = (
        FeatureSelectionAlgorithmFactory().get_name_to_classname_mapping()
    )
    __niaamlFeatureTransformAlgorithms = (
        FeatureTransformAlgorithmFactory().get_name_to_classname_mapping()
    )
    __niaamlClassifiers = ClassifierFactory().get_name_to_classname_mapping()
    __niaamlFitnessFunctions = FitnessFactory().get_name_to_classname_mapping()
    __niaamlEncoders = EncoderFactory().get_name_to_classname_mapping()
    __niaamlImputers = ImputerFactory().get_name_to_classname_mapping()
    __niapyAlgorithmsList = list(_algorithm_options().keys())
    __niaamlFeatureSelectionAlgorithmsList = list(
        __niaamlFeatureSelectionAlgorithms.keys()
    )
    __niaamlFeatureTransformAlgorithmsList = list(
        __niaamlFeatureTransformAlgorithms.keys()
    )
    __niaamlClassifiersList = list(__niaamlClassifiers.keys())
    __niaamlFitnessFunctionsList = list(__niaamlFitnessFunctions.keys())
    __niaamlEncodersList = list(__niaamlEncoders.keys())
    __niaamlImputersList = list(__niaamlImputers.keys())

    def __init__(self, parent, is_v1=False, *args, **kwargs):
        self.__niapyAlgorithmsList.sort()
        self.__niaamlFeatureSelectionAlgorithmsList.sort()
        self.__niaamlFeatureTransformAlgorithmsList.sort()
        self.__niaamlClassifiersList.sort()
        self.__is_v1 = is_v1
        super().__init__(parent, *args, **kwargs)

        fileLayout = QHBoxLayout(self._parent)

        selectFileBar = QHBoxLayout(self._parent)
        selectFileBar.setSpacing(5)
        selectFileBar.setContentsMargins(0, 5, 5, 5)
        fNameLine = QLineEdit(self._parent)
        fNameLine.setObjectName("csvFile")
        fNameLine.setPlaceholderText("Select a CSV dataset file...")
        fNameLine.setReadOnly(True)
        font = fNameLine.font()
        font.setPointSize(12)
        fNameLine.setFont(font)
        selectFileBar.addWidget(fNameLine)
        editBtn = self._createButton(
            None,
            self._editCSVFile,
            "editCSVButton",
            qta.icon("fa5.edit", color_off="white"),
        )
        editBtn.setEnabled(False)
        selectFileBar.addWidget(editBtn)
        selectFileBar.addWidget(
            self._createButton(
                None, self._openCSVFile, icon=qta.icon("fa5.file", color_off="white")
            )
        )

        checkBox = QCheckBox("CSV has header")
        checkBox.setObjectName("csv")
        checkBox.setFont(font)

        fileLayout.addItem(selectFileBar)
        fileLayout.addWidget(checkBox)

        encoders = self.__createComboBox(
            "Categorical features' encoder:", self.__niaamlEncodersList, "encoders"
        )
        imputers = self.__createComboBox(
            "Missing features' imputer:", self.__niaamlImputersList, "imputers"
        )

        hBoxLayout = QHBoxLayout(self._parent)
        hBoxLayout.setContentsMargins(0, 0, 0, 0)

        vBoxLayout = QVBoxLayout(self._parent)
        vBoxLayout.setContentsMargins(0, 34, 0, 0)

        h1BoxLayout = QHBoxLayout(self._parent)
        h1BoxLayout.setContentsMargins(0, 0, 0, 0)

        fsasBox = self.__createGridLayoutBox((0, 5, 3, 5), True)
        fsasList = self.__createListWidget([], None, "fsasList")
        fsasBox.addWidget(fsasList)
        h1BoxLayout.addItem(fsasBox)

        ftasBox = self.__createGridLayoutBox((3, 5, 3, 5), True)
        ftasList = self.__createListWidget([], None, "ftasList")
        ftasBox.addWidget(ftasList)
        h1BoxLayout.addItem(ftasBox)

        classifiers = self.__createGridLayoutBox((3, 5, 5, 5), True)
        classifiersList = self.__createListWidget([], None, "classifiersList")
        classifiers.addWidget(classifiersList)
        h1BoxLayout.addItem(classifiers)

        settingsBox = self.__createGridLayoutBox((0, 0, 5, 5), False, "transparent")
        settingsBox.setVerticalSpacing(10)

        optAlgosLabel = (
            "Optimization Algorithm (components selection):"
            if not self.__is_v1
            else "Optimization Algorithm:"
        )
        optAlgos = self.__createComboBox(
            optAlgosLabel, self.__niapyAlgorithmsList, "optAlgos"
        )

        optAlgosInner = self.__createComboBox(
            "Optimization Algorithm (parameter tuning) - same as first if not selected:",
            [*["None"], *self.__niapyAlgorithmsList],
            "optAlgosInner",
        )

        validator = QRegularExpressionValidator(QRegularExpression("[1-9][0-9]*"))

        popSizeLabel = (
            "Population size (components selection):"
            if not self.__is_v1
            else "Population size:"
        )
        popSize = self.__createTextInput(popSizeLabel, "popSize", validator)

        popSizeInner = self.__createTextInput(
            "Population size (parameter tuning):", "popSizeInner", validator
        )

        numEvalsLabel = (
            "Number of evaluations (components selection):"
            if not self.__is_v1
            else "Number of evaluations"
        )
        numEvals = self.__createTextInput(numEvalsLabel, "numEvals", validator)

        numEvalsInner = self.__createTextInput(
            "Number of evaluations (parameter tuning):", "numEvalsInner", validator
        )

        fitFuncs = self.__createComboBox(
            "Fitness Function:", self.__niaamlFitnessFunctionsList, "fitFuncs"
        )

        selectOutputFolderBar = QHBoxLayout(self._parent)
        selectOutputFolderBar.setSpacing(0)
        foNameLine = QLineEdit(self._parent)
        foNameLine.setObjectName("outputFolder")
        foNameLine.setPlaceholderText("Select pipeline output folder...")
        foNameLine.setReadOnly(True)
        font = foNameLine.font()
        font.setPointSize(12)
        foNameLine.setFont(font)
        selectOutputFolderBar.addWidget(foNameLine)
        selectOutputFolderBar.addWidget(
            self._createButton(
                None,
                self.__selectDirectory,
                icon=qta.icon("fa5.folder", color_off="white"),
            )
        )
        selectOutputFolderBar.setSpacing(5)

        settingsBox.addItem(optAlgos)
        if not self.__is_v1:
            settingsBox.addItem(optAlgosInner)
        settingsBox.addItem(popSize)
        if not self.__is_v1:
            settingsBox.addItem(popSizeInner)
        settingsBox.addItem(numEvals)
        if not self.__is_v1:
            settingsBox.addItem(numEvalsInner)
        settingsBox.addItem(fitFuncs)
        settingsBox.addItem(selectOutputFolderBar)

        confirmBar = QHBoxLayout(self._parent)
        confirmBar.setContentsMargins(5, 5, 5, 5)
        confirmBar.addStretch()
        confirmBar.addWidget(
            self._createButton("Start optimization", self.__runOptimize)
        )

        vBoxLayout.addItem(fileLayout)
        vBoxLayout.addItem(encoders)
        vBoxLayout.addItem(imputers)
        vBoxLayout.addItem(h1BoxLayout)
        vBoxLayout.addItem(settingsBox)
        vBoxLayout.addItem(confirmBar)

        exploreBox = self.__createGridLayoutBox((0, 0, 0, 0), False)
        exploreBox.addWidget(self.__createTabs(fsasList, ftasList, classifiersList))

        hBoxLayout.addItem(exploreBox)
        hBoxLayout.addItem(vBoxLayout)
        hBoxLayout.setStretchFactor(exploreBox, 1)
        hBoxLayout.setStretchFactor(vBoxLayout, 2)

        self.setLayout(hBoxLayout)

    def __createComboBox(self, label, items, name):
        comboBox = QVBoxLayout()
        comboBox.setSpacing(5)
        label = QLabel(label, self._parent)
        font = label.font()
        font.setPointSize(12)
        label.setFont(font)
        cb = QComboBox()
        cb.setObjectName(name)
        cb.setFont(font)
        for k in items:
            cb.addItem(k)
        comboBox.addWidget(label)
        comboBox.addWidget(cb)
        return comboBox

    def __createTextInput(self, label, name, validator=None):
        textBox = QHBoxLayout()
        textBox.setSpacing(5)
        label = QLabel(label, self._parent)
        font = label.font()
        font.setPointSize(12)
        label.setFont(font)
        tb = QLineEdit(self._parent)
        tb.setObjectName(name)
        tb.setFont(font)
        textBox.addWidget(label)
        textBox.addWidget(tb)

        if validator is not None:
            tb.setValidator(validator)

        return textBox

    def __createGridLayoutBox(
        self, tupleMargins, visibleBorder, background_color="#fff"
    ):
        layout = QGridLayout()
        layout.setContentsMargins(*tupleMargins)
        return layout

    def __createListWidget(self, items, targetBox=None, name=None):
        listWidget = ListWidgetCustom(items, targetBox, name)
        font = listWidget.font()
        font.setPointSize(12)
        listWidget.setFont(font)
        return listWidget

    def __createTabs(self, fsasList, ftasList, classifiersList):
        tabs = QTabWidget(self._parent)

        fsas = self.__createListWidget(
            self.__niaamlFeatureSelectionAlgorithmsList, fsasList
        )
        fsasList.setTarget(fsas)
        tabs.addTab(fsas, "Feature Selection Algorithms")

        ftas = self.__createListWidget(
            self.__niaamlFeatureTransformAlgorithmsList, ftasList
        )
        ftasList.setTarget(ftas)
        tabs.addTab(ftas, "Feature Transform Algorithms")

        clas = self.__createListWidget(self.__niaamlClassifiersList, classifiersList)
        classifiersList.setTarget(clas)
        tabs.addTab(clas, "Classifiers")

        font = tabs.font()
        font.setPointSize(10)
        tabs.setFont(font)
        tabs.setStyleSheet("QTabBar::tab { height: 40px; }")
        return tabs

    def __selectDirectory(self):
        fname = str(
            QFileDialog.getExistingDirectory(
                parent=self._parent, caption="Select Directory"
            )
        )
        self.findChild(QLineEdit, "outputFolder").setText(fname)

    def __runOptimize(self):
        err = ""

        csvSrc = self.findChild(QLineEdit, "csvFile").text()
        if self._isNoneOrWhiteSpace(csvSrc):
            err += "Select CSV dataset file.\n"

        encoderName = self.__niaamlEncoders[
            str(self.findChild(QComboBox, "encoders").currentText())
        ]
        imputerName = self.__niaamlImputers[
            str(self.findChild(QComboBox, "imputers").currentText())
        ]

        optAlgName = str(self.findChild(QComboBox, "optAlgos").currentText())

        if not self.__is_v1:
            optAlgInnerName = str(
                self.findChild(QComboBox, "optAlgosInner").currentText()
            )
            if optAlgInnerName == "None":
                optAlgInnerName = optAlgName

        popSize = self.findChild(QLineEdit, "popSize").text()
        if self._isNoneOrWhiteSpace(popSize):
            err += "Select population size.\n"
        else:
            try:
                popSize = int(popSize)
            except ValueError:
                err += "Invalid population size value.\n"

        if not self.__is_v1:
            popSizeInner = self.findChild(QLineEdit, "popSizeInner").text()
            if self._isNoneOrWhiteSpace(popSizeInner):
                err += "Select inner population size.\n"
            else:
                try:
                    popSizeInner = int(popSizeInner)
                except ValueError:
                    err += "Invalid inner population size value.\n"

        numEvals = self.findChild(QLineEdit, "numEvals").text()
        if self._isNoneOrWhiteSpace(numEvals):
            err += "Select number of evaluations.\n"
        else:
            try:
                numEvals = int(numEvals)
            except ValueError:
                err += "Invalid number of evaluations.\n"

        if not self.__is_v1:
            numEvalsInner = self.findChild(QLineEdit, "numEvalsInner").text()
            if self._isNoneOrWhiteSpace(numEvalsInner):
                err += "Select number of inner evaluations.\n"
            else:
                try:
                    numEvalsInner = int(numEvalsInner)
                except ValueError:
                    err += "Invalid number of inner evaluations.\n"

        fsasList = self.findChild(ListWidgetCustom, "fsasList")
        fsas = [
            self.__niaamlFeatureSelectionAlgorithms[fsasList.item(i).text()]
            for i in range(fsasList.count())
        ]

        ftasList = self.findChild(ListWidgetCustom, "ftasList")
        ftas = [
            self.__niaamlFeatureTransformAlgorithms[ftasList.item(i).text()]
            for i in range(ftasList.count())
        ]

        clsList = self.findChild(ListWidgetCustom, "classifiersList")
        classifiers = [
            self.__niaamlClassifiers[clsList.item(i).text()]
            for i in range(clsList.count())
        ]
        if len(classifiers) == 0:
            err += "Select at least one classifier.\n"

        fitnessFunctionName = self.__niaamlFitnessFunctions[
            str(self.findChild(QComboBox, "fitFuncs").currentText())
        ]

        outputFolder = self.findChild(QLineEdit, "outputFolder").text()
        if self._isNoneOrWhiteSpace(outputFolder):
            err += "Select an output directory.\n"

        if not self._isNoneOrWhiteSpace(err):
            self._parent.errorMessage.setText(err)
            self._parent.errorMessage.show()
            return
        pipelineSettings = {
            "csvSrc": csvSrc,
            "csvHasHeader": self.findChild(QCheckBox, "csv").isChecked(),
            "encoder": encoderName,
            "imputer": imputerName,
            "optAlgName": optAlgName,
            "optAlgInnerName": optAlgInnerName if not self.__is_v1 else None,
            "popSize": popSize,
            "popSizeInner": popSizeInner if not self.__is_v1 else None,
            "numEvals": numEvals,
            "numEvalsInner": numEvalsInner if not self.__is_v1 else None,
            "fsas": ", ".join(fsas),
            "ftas": ", ".join(ftas),
            "classifiers": ", ".join(classifiers),
            "fitnessFunctionName": fitnessFunctionName,
            "outputFolder": outputFolder,
        }
        if not self.__is_v1:
            self._processWindow = ProcessWindow(
                self._parent,
                ProcessWindowData(
                    True,
                    csvSrc,
                    self.findChild(QCheckBox, "csv").isChecked(),
                    encoderName,
                    imputerName,
                    optAlgName,
                    optAlgInnerName,
                    popSize,
                    popSizeInner,
                    numEvals,
                    numEvalsInner,
                    fsas,
                    ftas,
                    classifiers,
                    fitnessFunctionName,
                    outputFolder,
                ),
                pipelineSettings
            )
        else:
            self._processWindow = ProcessWindow(
                self._parent,
                ProcessWindowData(
                    "v1",
                    csvSrc,
                    self.findChild(QCheckBox, "csv").isChecked(),
                    encoderName,
                    imputerName,
                    optAlgName=optAlgName,
                    popSize=popSize,
                    numEvals=numEvals,
                    fsas=fsas,
                    ftas=ftas,
                    classifiers=classifiers,
                    fitnessFunctionName=fitnessFunctionName,
                    outputFolder=outputFolder,
                ),
                pipelineSettings
            )

        #self._parent.setResultsView(results_data,results_data, pipelineSettings)
        self._processWindow.show()



from PyQt6.QtWidgets import QWidget, QVBoxLayout, QLabel, QTableWidget, QTableWidgetItem, QPushButton, QSizePolicy
from PyQt6.QtCore import Qt
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.figure import Figure

class ResultsWidget(QWidget):
    def __init__(self, parent=None, results_data=None, pipeline_settings=None):
        super().__init__(parent)
        self.setObjectName("resultsView")
        self._parent = parent
        self.results_data = results_data or {}
        self.pipeline_settings = pipeline_settings or {}
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Maximum)
        layout = QVBoxLayout()

        title = QLabel("Optimization Results")
        title.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(title)
        
        # stats table
        table = QTableWidget(4, 2)
        table.setHorizontalHeaderLabels(["Metric", "Value"])
        table.setFixedHeight(table.verticalHeader().length() + table.horizontalHeader().height())
        # table.resizeColumnsToContents()
        # table.resizeRowsToContents()

        table.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        table.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)        
        total_width = sum([table.columnWidth(i) for i in range(table.columnCount())])
        total_width += table.verticalHeader().width()
        total_width += 2 * table.frameWidth()
        table.setMaximumWidth(total_width)
        table.setStyleSheet("""
            QTableWidget {
                background-color: #005f85;
                color: white;
                border: 1px solid #001D85;
                gridline-color: #001D85;
            }
            QHeaderView::section {
                background-color: #001D85;
                color: white;
                font-weight: bold;
            }
            QTableWidget::item:selected {
                background-color: #008569;
                color: white;
            }
        """)
        metrics = [
            ("Accuracy", results_data.get("accuracy", 0.0)),
            ("Precision", results_data.get("precision", 0.0)),
            ("F1-score", results_data.get("f1_score", 0.0)),
            ("Cohen's Kappa", results_data.get("kappa", 0.0)),
        ]
        for row, (label, value) in enumerate(metrics):
            table.setItem(row, 0, QTableWidgetItem(label))
            table.setItem(row, 1, QTableWidgetItem(f"{value:.4f}"))

        layout.addWidget(table)
        # graph
        fig = Figure()
        fig = Figure(facecolor="#005f85")
        
        canvas = FigureCanvas(fig)
        canvas.setObjectName("graphCanvas")
          
        ax = fig.add_subplot(111)
        ax.set_facecolor("#005f85")  
        ax.bar(["Accuracy", "Precision", "F1", "Kappa"], [0.4634, 0.4478, 0.4339, 0.174])
        ax.tick_params(colors="white")
        ax.spines['bottom'].set_color('white')
        ax.spines['left'].set_color('white')
        ax.yaxis.label.set_color('white')
        ax.xaxis.label.set_color('white')
        
        layout.addWidget(canvas)
        
        # pipeline settings table
        if pipeline_settings:
            pipeline_table = QTableWidget(len(pipeline_settings), 2)
            pipeline_table.setHorizontalHeaderLabels(["Parameter", "Value"])
            for row, (key, value) in enumerate(pipeline_settings.items()):
                pipeline_table.setItem(row, 0, QTableWidgetItem(str(key)))
                pipeline_table.setItem(row, 1, QTableWidgetItem(str(value)))

            pipeline_table.resizeColumnsToContents()  
            pipeline_table.resizeRowsToContents()

            pipeline_table.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
            pipeline_table.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)

            total_width = sum([pipeline_table.columnWidth(i) for i in range(pipeline_table.columnCount())])
            total_width += pipeline_table.verticalHeader().width()
            total_width += 2 * pipeline_table.frameWidth()
            pipeline_table.setMaximumWidth(total_width)

            pipeline_table.setStyleSheet("""
                QTableWidget {
                    background-color: #005f85;
                    color: white;
                    border: 1px solid #001D85;
                    gridline-color: #001D85;
                }
                QHeaderView::section {
                    background-color: #001D85;
                    color: white;
                    font-weight: bold;
                }
                QTableWidget::item:selected {
                    background-color: #008569;
                    color: white;
                }
            """)

            layout.addWidget(pipeline_table)
        
        back_btn = QPushButton("Back to main view")
        back_btn.clicked.connect(self.__goBack)
        layout.addWidget(back_btn)

        self.setLayout(layout)

    def __goBack(self):
        if self._parent:
            from niaaml_gui.widgets import OptimizationWidget
            self._parent.setCentralWidget(OptimizationWidget(self._parent))



import sys
from PyQt6.QtWidgets import QApplication, QMainWindow
from niaaml_gui.widgets.pipeline_canvas import PipelineCanvas

class CanvasTestWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("Pipeline Canvas Test")
        self.setGeometry(100, 100, 1200, 800)

        self.canvas = PipelineCanvas(self)
        self.setCentralWidget(self.canvas)

if __name__ == '__main__':
    app = QApplication(sys.argv)
    window = CanvasTestWindow()
    window.show()
    sys.exit(app.exec())



import traceback
from pathlib import Path
from typing import Callable, Sequence
from niaaml import PipelineOptimizer
from niaaml.data import CSVDataReader

def run_pipeline(
    *,
    csv_path: str,
    has_header: bool = True,
    contains_classes: bool = True,
    ignore_cols: list[int] | None = None,
    fitness_name: str,
    pop_size: int,
    inner_pop: int,
    evals: int,
    inner_evals: int,
    opt_alg: str,
    classifiers: Sequence[str],
    fs_algorithms: Sequence[str],
    ft_algorithms: Sequence[str],
    log_fn: Callable[[str], None] | None = None,
    save_path: Path | str | None = None,
):
    log = log_fn or (lambda *_: None)

    try:
        data_reader = CSVDataReader(
            src=csv_path,
            contains_classes=contains_classes,
            has_header=has_header,
        )

        pipeline_optimizer = PipelineOptimizer(
            data=data_reader,
            classifiers=list(classifiers) or ["RandomForest"],
            feature_selection_algorithms=list(fs_algorithms) or ["VarianceThreshold"],
            feature_transform_algorithms=list(ft_algorithms) or ["StandardScaler"],
            log=True
        )

        pipeline = pipeline_optimizer.run(
            fitness_name,
            pop_size, inner_pop,
            evals, inner_evals,
            opt_alg, opt_alg,
        )

        if pipeline is None:
            log(" pipeline_optimizer.run() returned None.")
            return None

        if save_path:
            save_dir = Path(save_path).resolve()
            save_dir.mkdir(parents=True, exist_ok=True)  

            target_ppln = save_dir / "NiaAML-GUI.ppln"
            target_txt  = save_dir / "NiaAML-GUI.txt"

            pipeline.export(target_ppln)
            pipeline.export_text(str(target_txt))

        return pipeline

    except Exception:
        traceback.print_exc()
        return None



from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QVBoxLayout, QLabel, QPushButton
import webbrowser

from niaaml_gui.widgets.base_main_widget import BaseMainWidget


class HelpLicenseWidget(BaseMainWidget):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        vBoxLayout = QVBoxLayout(self._parent)
        vBoxLayout.setAlignment(Qt.AlignmentFlag.AlignTop)

        license_title = QLabel("License")
        font = QFont()
        font.setPointSize(20)
        license_title.setFont(font)

        vBoxLayout.addWidget(license_title)

        contentFont = QFont()
        contentFont.setPointSize(12)

        license_content = QLabel("This package is distributed under the MIT License. "
                                    "This license can be found online at http://www.opensource.org/licenses/MIT.")
        license_content.setFont(contentFont)
        license_content.setStyleSheet("""
            QLabel {
                padding-bottom: 12px;
            }
        """)

        button = QPushButton('Open license')
        button.clicked.connect(self.open_license_link)
        button.setFont(contentFont)

        vBoxLayout.addWidget(license_content)
        vBoxLayout.addWidget(button)

        self.setLayout(vBoxLayout)

    def open_license_link(self):
        # Define the URL you want to open
        url = 'http://www.opensource.org/licenses/MIT'

        # Open the URL in the default web browser
        webbrowser.open(url)



import math
from PyQt6.QtWidgets import QGraphicsPathItem, QGraphicsPolygonItem
from PyQt6.QtGui import QPen, QColor, QPainterPath, QBrush, QPolygonF
from PyQt6.QtCore import QPointF, Qt


class ConnectionLine(QGraphicsPathItem):
    def __init__(self, source_block, target_block):
        super().__init__()
        self.source_block = source_block
        self.target_block = target_block

        self.setZValue(100)
        self.setPen(QPen(QColor("black"), 2))
        self.setFlags(
            QGraphicsPathItem.GraphicsItemFlag.ItemIsSelectable
        )

        self.arrow_shape = QPolygonF([
            QPointF(0, 0),
            QPointF(-12, -6),
            QPointF(-12, 6)
        ])

        self.arrow_item = QGraphicsPolygonItem(self.arrow_shape, self)
        self.arrow_item.setBrush(QBrush(QColor("black")))
        self.arrow_item.setPen(QPen(Qt.PenStyle.NoPen))
        self.arrow_item.setZValue(101)
        self.arrow_item.setTransformOriginPoint(QPointF(0, 0))

        self.update_path()

    def mousePressEvent(self, event):
        self.setSelected(True)
        super().mousePressEvent(event)

    def update_path(self):
        start = self.source_block.output_point()
        end = self.target_block.input_point()

        path = QPainterPath()
        path.moveTo(start)

        pre_end = QPointF(end.x() - 20, end.y())

        dx = (pre_end.x() - start.x()) * 0.5
        cp1 = QPointF(start.x() + dx, start.y())
        cp2 = QPointF(pre_end.x() - dx, pre_end.y())
        path.cubicTo(cp1, cp2, pre_end)

        path.lineTo(end)
        self.setPath(path)

        arrow_size = 12
        arrow = QPolygonF([
            QPointF(0, 0),
            QPointF(-arrow_size, -arrow_size / 2),
            QPointF(-arrow_size, arrow_size / 2)
        ])

        if hasattr(self, "arrow_item") and self.arrow_item:
            if self.arrow_item.scene():
                self.arrow_item.scene().removeItem(self.arrow_item)
            self.arrow_item = None

        self.arrow_item = QGraphicsPolygonItem(arrow, self)
        self.arrow_item.setBrush(QBrush(QColor("black")))
        self.arrow_item.setPen(QPen(Qt.PenStyle.NoPen))
        self.arrow_item.setZValue(101)

        self.arrow_item.setRotation(0)
        self.arrow_item.setPos(end)





from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QListWidget


class ListWidgetCustom(QListWidget):
    def __init__(self, items, target, name, *args, **kwargs):
        super(QListWidget, self).__init__(*args, **kwargs)
        self.__items = items
        self.__target = target

        for k in self.__items:
            self.addItem(k)

        self.sortItems(Qt.SortOrder.AscendingOrder)
        self.itemClicked.connect(self.__clicked)
        self.setObjectName(name)

    def setTarget(self, value):
        self.__target = value

    def __clicked(self, item):
        if self.__target is not None:
            self.__target.addItem(item.text())
            self.takeItem(self.row(item))
            self.__target.sortItems(Qt.SortOrder.AscendingOrder)



from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QProgressBar


class ProgressBar(QProgressBar):
    def __init__(self):
        super(ProgressBar, self).__init__()
        self.setTextVisible(True)
        self.setMaximum(100)
        self.setAlignment(Qt.AlignmentFlag.AlignCenter)



from PyQt6.QtWidgets import (
    QMainWindow, QVBoxLayout, QWidget, QPlainTextEdit,
    QPushButton, QHBoxLayout
)
from PyQt6.QtCore import QSize, Qt
from niaaml_gui.progress_bar import ProgressBar
from niaaml_gui.windows.threads import OptimizeThread
from niaaml_gui.windows.threads.pipeline_runner_thread import PipelineRunnerThread
from pyqt_feedback_flow.feedback import TextFeedback, AnimationType, AnimationDirection
import copy, re, os


class ProcessWindow(QMainWindow):
    def __init__(self, parent, data, pipelineSettings):
        super().__init__(parent)
        self.setMinimumSize(QSize(640, 480))
        self._parent = parent

        centralWidget = QWidget(self)
        layout = QVBoxLayout(centralWidget)
        layout.setAlignment(Qt.AlignmentFlag.AlignTop)

        self.__progressBar = ProgressBar()
        layout.addWidget(self.__progressBar)
        self.__pipelineSettings = pipelineSettings

        self.__textArea = QPlainTextEdit(parent=self)
        self.__textArea.setReadOnly(True)
        layout.addWidget(self.__textArea)

        confirmBar = QHBoxLayout(self)
        confirmBar.addStretch()

        self.__btn = QPushButton(self)
        self.__btn.setText("Cancel")
        font = self.__btn.font()
        font.setPointSize(12)
        self.__btn.setFont(font)
        self.__btn.clicked.connect(self.cancelClose)
        confirmBar.addWidget(self.__btn)

        layout.addItem(confirmBar)
        centralWidget.setLayout(layout)
        self.setCentralWidget(centralWidget)

        self.__data = copy.deepcopy(data)

        if self.__data.isOptimization is True or self.__data.isOptimization == "v1":
            self.__progressBar.setMaximum(100)
            self.__currentEvals = 0
            self.__totalEvals = (
                data.numEvals * data.numEvalsInner
                if data.isOptimization is True
                else data.numEvals
            )

            optimizer = OptimizeThread(self.__data)
            optimizer.optimized.connect(self.onOptimizationComplete)
            optimizer.progress.connect(self.onOptimizationProgress)

            self.__runningThread = optimizer
            optimizer.start()
            self.__textArea.appendPlainText("Pipeline optimization running...\n")
            self.show_toast("Pipeline optimization started 🚀")

        else:
            self.__progressBar.setMaximum(100)
            self.__progressBar.setValue(0)
            self.__progressBar.setTextVisible(True)
            self.__progressBar.show()

            self.__currentEvals = 0
            self.__totalEvals = (
                int(self.__data.numEvals or 1) * int(self.__data.numEvalsInner or 1)
            )

            runner = PipelineRunnerThread(self.__data)
            runner.ran.connect(self.onRunComplete)
            runner.progress.connect(self.onOptimizationProgress)
            self.__runningThread = runner
            runner.start()
            self.__textArea.appendPlainText("Pipeline running...\n")
            self.show_toast("Pipeline started 🏃‍♂️")

    def show_toast(self, message: str, direction=AnimationDirection.UP, animation=AnimationType.VERTICAL):
        toast = TextFeedback(message)
        toast.label.setStyleSheet("""
            background-color: #007199;
            color: white;
            font-size: 16pt;
            padding: 15px;
            border: 2px solid #001d85;
            border-radius: 12px;
            font-family: Segoe UI;
        """)
        toast.show(animation, direction, 3000)


    def cancelClose(self):
        self.close()
        try:
            self.__runningThread.terminate()
        except BaseException as e:
            print("terminate() failed:", e)

    def onOptimizationComplete(self, data):
        self.__progressBar.setValue(100)
        self.__textArea.appendPlainText(data + "\n")
        self.__textArea.appendPlainText("Pipeline optimization complete.")
        self.show_toast("Optimization complete ✅")

        result_file_path = os.path.join(self.__data.outputFolder, "NiaAML-GUI.txt")
        if os.path.exists(result_file_path):
            self.__textArea.appendPlainText("\n--- Results file content ---")
            with open(result_file_path, "r", encoding="utf-8") as f:
                content = f.read()
                self.__textArea.appendPlainText(content)
        else:
            self.__textArea.appendPlainText("No result file found at: " + result_file_path)

        self.__textArea.appendPlainText("Results exported to: " + self.__data.outputFolder)
        self.__btn.setText("Close")

        results_data = parse_optimization_output(data)
        self._parent.setResultsView(results_data, self.__pipelineSettings)

    def onOptimizationProgress(self, data):
        if any(kw in data for kw in ("Evaluation", "Generation", "Iteration")):
            self.__currentEvals += 1
            if self.__totalEvals == 0:
                self.__totalEvals = 1
            val = int((self.__currentEvals / self.__totalEvals) * 100)
            self.__progressBar.setValue(val)

    def onRunComplete(self, result):
        self.__progressBar.setMaximum(100)
        self.__progressBar.setValue(100)

        self.__textArea.appendPlainText("Predictions: " + result + "\n")

        result_file_path = os.path.join(self.__data.outputFolder, "NiaAML-GUI.txt")
        if os.path.exists(result_file_path):
            with open(result_file_path, "r", encoding="utf-8") as file:
                content = file.read()
                self.__textArea.appendPlainText("\n--- Results from NiaAML-GUI.txt ---\n")
                self.__textArea.appendPlainText(content)
        else:
            self.__textArea.appendPlainText("\n⚠️ File 'NiaAML-GUI.txt' not found in output folder.")

        self.__textArea.appendPlainText("Pipeline run complete.")
        self.__btn.setText("Close")
        self.show_toast("Pipeline finished 🎉")



def parse_optimization_output(text: str):
    results_data = {}

    accuracy = re.search(r"Accuracy:\s*([0-9.]+)", text)
    precision = re.search(r"Precision:\s*([0-9.]+)", text)
    f1_score = re.search(r"F1[-\s]?score:\s*([0-9.]+)", text)
    kappa = re.search(r"(?:Cohen's\s+kappa|Kappa):\s*([0-9.]+)", text)

    results_data["accuracy"] = float(accuracy.group(1)) if accuracy else 0.0
    results_data["precision"] = float(precision.group(1)) if precision else 0.0
    results_data["f1_score"] = float(f1_score.group(1)) if f1_score else 0.0
    results_data["kappa"] = float(kappa.group(1)) if kappa else 0.0

    return results_data



from PyQt6.QtWidgets import (
    QWidget, QVBoxLayout, QLabel, QListWidget, QListWidgetItem, QHBoxLayout,
    QFileDialog, QToolBox
)
from PyQt6.QtCore import Qt, QMimeData
from PyQt6.QtGui import QDrag, QPixmap
import os


class ComponentSidebar(QWidget):
    """Verzija, kjer je vsak zavihek prikazan eden pod drugim z uporabo QToolBox.
    Logika metod in MIME ostane nespremenjena.
    """

    def __init__(self, canvas):
        super().__init__()
        self.canvas = canvas
        self._base_width = 220
        self._item_height = 30

        root = QVBoxLayout(self)
        root.setAlignment(Qt.AlignmentFlag.AlignTop)
        root.setContentsMargins(0, 0, 0, 0)

        title = QLabel("Drag Components:")
        title.setStyleSheet("color: white; font-weight: bold; padding: 5px;")
        title.setAlignment(Qt.AlignmentFlag.AlignCenter)
        root.addWidget(title)
        root.addSpacing(6)

        from PyQt6.QtWidgets import QToolBox
        self.tabs = QToolBox(self)
        root.addWidget(self.tabs)

        self.listWidget = None

        categories = {
            "Data": {
                "Select CSV File": "file.png",
                "Pipeline Output Folder": "folder.png",
            },
            "Pre‑processing": {
                "Categorical Encoder": "encode.png",
                "Missing Imputer": "imputation.png",
            },
            "Feature Engineering": {
                "Feature Selection": "feature_selection.png",
                "Feature Transform": "feature_transformation.png",
            },
            "Modeling": {
                "Classifier": "classifier.png",
                "Optimization Algorithm (Selection)": "optimization_algorithm.png",
                "Optimization Algorithm (Tuning)": "decision_tree.png",
            },
            "Fitness": {
                "Fitness Function": "fitness.png",
            },
            "Number of evaluations": {
                "Number of Evaluations (Component Selection)": "numbers.png",
                "Number of Evaluations (Parameter Tuning)": "numbers.png",
            },
            "Population Size": {
                "Population Size (Components Selection)": "numbers.png",
                "Population Size (Parameter Tuning)": "numbers.png",
            },
        }
        
        icon_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), "resources", "icons")

        for cat_name, items in categories.items():
            lst = QListWidget()
            lst.setSpacing(2)
            lst.setDragEnabled(True)
            lst.mouseMoveEvent = self._wrapped_start_drag
            #lst.itemClicked.connect(self.handle_click)

            for label, icon_file in items.items():
                icon_path = os.path.join(icon_dir, icon_file)
                if not os.path.exists(icon_path):
                    icon_path = os.path.join(icon_dir, "placeholder.png")

                item = QListWidgetItem()
                item.setData(Qt.ItemDataRole.UserRole, label)

                widget = QWidget()
                hlyt = QHBoxLayout(widget)
                hlyt.setContentsMargins(5, 2, 5, 2)

                icon_lbl = QLabel()
                icon_lbl.setPixmap(
                    QPixmap(icon_path).scaled(24, 24, Qt.AspectRatioMode.KeepAspectRatio,
                                              Qt.TransformationMode.SmoothTransformation)
                )
                hlyt.addWidget(icon_lbl)

                text_lbl = QLabel(label)
                text_lbl.setWordWrap(True)
                text_lbl.setStyleSheet("color: white;")
                text_lbl.setAlignment(Qt.AlignmentFlag.AlignRight | Qt.AlignmentFlag.AlignVCenter)
                hlyt.addWidget(text_lbl)

                widget.setLayout(hlyt)
                item.setSizeHint(widget.sizeHint())
                lst.addItem(item)
                lst.setItemWidget(item, widget)

            self.tabs.addItem(lst, cat_name)

        self.listWidget = self.tabs.currentWidget()
        self.tabs.currentChanged.connect(self._on_tab_changed)
        self._adjust_width()

    def _on_tab_changed(self, idx: int):
        self.listWidget = self.tabs.widget(idx)
        self._adjust_width()

    def _wrapped_start_drag(self, event):
        item = self.listWidget.itemAt(event.pos()) if self.listWidget else None
        if not item:
            return
        text = item.data(Qt.ItemDataRole.UserRole)
        mime = QMimeData()
        mime.setText(text)
        drag = QDrag(self.listWidget)
        drag.setMimeData(mime)
        drag.exec()

    def handle_click(self, item):
        label = item.data(Qt.ItemDataRole.UserRole)
        if label == "Select CSV File":
            path, _ = QFileDialog.getOpenFileName(
                self, "Select CSV File", "", "CSV Files (*.csv);;All Files (*)"
            )
            if path:
                self.canvas.add_config_block("CSV File", [], value=path)
        elif label == "Pipeline Output Folder":
            path = QFileDialog.getExistingDirectory(self, "Select Output Folder", "")
            if path:
                self.canvas.add_config_block("Output Folder", [], value=path)

    def _adjust_width(self):
        if not self.listWidget:
            return
        self.setFixedWidth(self._base_width)

    def dropEvent(self, event):
        if event.mimeData().hasText():
            label = event.mimeData().text()
            drop_pos = self.mapToScene(event.position().toPoint())
            self.add_config_block(label, x=drop_pos.x(), y=drop_pos.y())
            event.acceptProposedAction()

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dragMoveEvent(self, event):
        event.acceptProposedAction()



from pathlib import Path
import sys
import os
from PyQt6.QtWidgets import QMainWindow, QMessageBox, QApplication, QWidget, QHBoxLayout
from PyQt6.QtGui import QAction
from PyQt6.QtCore import QSize
from niaaml_gui.widgets.pipeline_canvas import PipelineCanvas
from niaaml_gui.widgets.sidebar import ComponentSidebar
from niaaml_gui.windows.process_window import ProcessWindow
from niaaml_gui.windows.threads.optimize_thread import OptimizeThread
from niaaml_gui.widgets.pipeline_controls import PipelineControlsWidget
from niaaml_gui.process_window_data import ProcessWindowData
from niaaml.classifiers import ClassifierFactory
from niaaml.preprocessing.feature_selection import FeatureSelectionAlgorithmFactory
from niaaml.preprocessing.feature_transform import FeatureTransformAlgorithmFactory
from niaaml.preprocessing.feature_selection import FeatureSelectionAlgorithmFactory
from niaaml.preprocessing.feature_transform import FeatureTransformAlgorithmFactory
from niaaml.fitness import FitnessFactory
from niaaml.preprocessing.encoding import EncoderFactory
from niaaml.preprocessing.imputation import ImputerFactory
from niapy.util.factory import _algorithm_options
import pandas as pd
from niaaml_gui.utils.pipeline_runner import run_pipeline as nia_run


class MainAppWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.setMinimumSize(QSize(1024, 768))
        self.setWindowTitle("NiaAML - GUI")

        # Menu
        menuBar = self.menuBar()
        menuBar.setNativeMenuBar(False)
        fileMenu = menuBar.addMenu("&File")
        helpMenu = menuBar.addMenu("&Help")

        exitAction = QAction(text="Exit", parent=self)
        exitAction.triggered.connect(QApplication.quit)
        fileMenu.addAction(exitAction)

        # Layout
        centralWidget = QWidget(self)
        mainLayout = QHBoxLayout(centralWidget)

        self.pipelineCanvas = PipelineCanvas()
        self.sidebar = ComponentSidebar(self.pipelineCanvas)
        self.controls = PipelineControlsWidget()

        self.controls.runClicked.connect(self.run_pipeline)
        self.controls.resetClicked.connect(self.reset_pipeline)
        self.pipelineCanvas.pipelineStateChanged.connect(self.validate_pipeline_ready)
        self._update_run_button_state()  

        mainLayout.addWidget(self.sidebar)
        mainLayout.addWidget(self.pipelineCanvas)
        mainLayout.addWidget(self.controls)

        centralWidget.setLayout(mainLayout)
        self.setCentralWidget(centralWidget)

        # Napake
        self.errorMessage = QMessageBox()
        self.errorMessage.setIcon(QMessageBox.Icon.Critical)
        self.errorMessage.setWindowTitle("Error")
        self.errorMessage.setStandardButtons(QMessageBox.StandardButton.Ok)

    def run_pipeline(self):
        
        invalid_blocks = []

        for block, info in self.pipelineCanvas.block_data.items():
            label = info.get("label", "")
            if hasattr(block, "get_value"):
                value = block.get_value()
            elif hasattr(block, "dropdown"):
                value = block.dropdown.currentText()
            else:
                value = info.get("path") or getattr(block, "value", None)

            if not value or (isinstance(value, str) and not value.strip()):
                invalid_blocks.append(label)

        if invalid_blocks:
            msg = "Pipeline isn't set up correctly. Missing components:\n\n" + "\n".join(f"- {label}" for label in invalid_blocks)
            self.errorMessage.setText(msg)
            self.errorMessage.exec()
            return
        
        blocks = self.pipelineCanvas.block_data
        if not blocks:
            self.errorMessage.setText("Pipeline is empty!")
            self.errorMessage.exec()
            return

        data = {}
        has_header = True 
        encoder_map = EncoderFactory().get_name_to_classname_mapping()
        imputer_map = ImputerFactory().get_name_to_classname_mapping()
        fitness_map = FitnessFactory().get_name_to_classname_mapping()
        opt_algo_map = _algorithm_options()
        fs_map = FeatureSelectionAlgorithmFactory().get_name_to_classname_mapping()
        ft_map = FeatureTransformAlgorithmFactory().get_name_to_classname_mapping()
        clf_map = ClassifierFactory().get_name_to_classname_mapping()

        for block, info in blocks.items():
            label = info["label"]

            if hasattr(block, "get_value"):
                value = block.get_value()
            elif hasattr(block, "dropdown"):
                value = block.dropdown.currentText()
            else:
                value = info.get("path") or getattr(block, "value", None)
 
            if label == "Categorical Encoder":
                value = encoder_map.get(value.strip(), value)
            elif label == "Missing Imputer":
                value = imputer_map.get(value.strip(), value)
            elif label == "Fitness Function":
                value = fitness_map.get(value.strip(), value)
            elif label in [
                "Optimization Algorithm (Selection)",
                "Optimization Algorithm (Tuning)",
            ]:
                rev = {v: k for k, v in opt_algo_map.items()}
                value = rev.get(value.strip(), value)
            elif label == "Feature Selection":
                value = "\n".join(
                    fs_map.get(v.strip(), v) for v in value.split("\n") if v.strip()
                )
            elif label == "Feature Transform":
                value = "\n".join(
                    ft_map.get(v.strip(), v) for v in value.split("\n") if v.strip()
                )
            elif label == "Classifier":
                value = "\n".join(
                    clf_map.get(v.strip(), v) for v in value.split("\n") if v.strip()
                )
            if label == "Select CSV File":
                if hasattr(block, "checkbox"):
                    has_header = block.checkbox.isChecked()
                data["csvSrc"] = value
            else:
                data[label] = value
                
        data["csvHasHeader"] = has_header
        self.currentPipelineData = ProcessWindowData.from_dict(data)


        
        self._processWindow = ProcessWindow(
            parent=self,
            data=self.currentPipelineData,
            pipelineSettings={                 
                "classifiers": self.currentPipelineData.classifiers.split("\n"),
                "fs_algorithms": self.currentPipelineData.fsas.split("\n"),
                "ft_algorithms": self.currentPipelineData.ftas.split("\n"),
            },
        )
        self._processWindow.show()        
    
        # nia_run(
        #     csv_path=self.currentPipelineData.csvSrc,
        #     has_header=self.currentPipelineData.csvHasHeader,
        #     contains_classes=True,
        #     ignore_cols=[],
        #     fitness_name=self.currentPipelineData.fitnessFunctionName,
        #     pop_size=int(self.currentPipelineData.popSize or 20),
        #     inner_pop=int(self.currentPipelineData.popSizeInner or 20),
        #     evals=int(self.currentPipelineData.numEvals or 200),
        #     inner_evals=int(self.currentPipelineData.numEvalsInner or 200),
        #     opt_alg=self.currentPipelineData.optAlgName or "BatAlgorithm",
        #     classifiers=self.currentPipelineData.classifiers.split("\n"),
        #     fs_algorithms=self.currentPipelineData.fsas.split("\n"),
        #     ft_algorithms=self.currentPipelineData.ftas.split("\n"),
        #     log_fn=lambda *_: None,
        #     save_path=Path(self.currentPipelineData.outputFolder, "niaamlGUIoutput"),
        # )

    def reset_pipeline(self):
        self.pipelineCanvas.scene.clear()
        self.pipelineCanvas.block_data.clear()
        self.currentPipelineData = None

    def show_optimization_result(self, result_text):
        QMessageBox.information(self, "Pipeline Finished", result_text)
        
    def validate_pipeline_ready(self):
        all_valid = True
        for block, info in self.pipelineCanvas.block_data.items():
            label = info.get("label", "")
            if hasattr(block, "get_value"):
                value = block.get_value()
            elif hasattr(block, "dropdown"):
                value = block.dropdown.currentText()
            else:
                value = info.get("path") or getattr(block, "value", None)
                            
            if not value or (isinstance(value, str) and not value.strip()):
                all_valid = False
                break

        self.controls.setRunEnabled(all_valid)

    def _update_run_button_state(self):
        ready = self.pipelineCanvas.is_pipeline_ready()
        self.controls.run_button.setEnabled(ready)

def run():
    app = QApplication(sys.argv)
    app.setStyle("Fusion")

    style_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "styles.qss")
    if os.path.exists(style_path):
        with open(style_path, "r") as f:
            style_sheet = f.read()
            app.setStyleSheet(style_sheet)
            print("styles.qss loaded and applied.")
    else:
        print("styles.qss NOT found!")

    mainWin = MainAppWindow()
    mainWin.showMaximized()
    sys.exit(app.exec())


if __name__ == "__main__":
    run()



from niaaml_gui.widgets.list_widget_custom import ListWidgetCustom
from niaaml_gui.widgets.optimization_widget import OptimizationWidget
from niaaml_gui.widgets.use_pipeline_widget import UsePipelineWidget

__all__ = ["ListWidgetCustom", "OptimizationWidget", "UsePipelineWidget"]



from PyQt6.QtWidgets import QWidget, QPushButton, QLineEdit, QFileDialog
from niaaml_gui.windows import CSVEditorWindow
from PyQt6.QtCore import QSize


class BaseMainWidget(QWidget):
    def __init__(self, parent, *args, **kwargs):
        super(QWidget, self).__init__(*args, **kwargs)
        self._parent = parent
        self._processWindow = None
        self.__csvEditWindow = None

    def _createButton(self, text, callback=None, objectName=None, icon=None):
        btn = QPushButton(self._parent)
        btn.setText(text)
        font = btn.font()
        font.setPointSize(12)
        btn.setFont(font)

        if callback is not None:
            btn.clicked.connect(callback)

        if objectName is not None:
            btn.setObjectName(objectName)

        if icon is not None:
            btn.setIcon(icon)
            btn.setIconSize(QSize(21, 21))

        return btn

    def _isNoneOrWhiteSpace(self, text):
        if text is None or text.isspace() or len(text) == 0:
            return True

    def _openCSVFile(self):
        fname = QFileDialog.getOpenFileName(
            parent=self._parent, caption="Select CSV File", filter="CSV files (*.csv)"
        )
        self.findChild(QLineEdit, "csvFile").setText(fname[0])
        self.findChild(QPushButton, "editCSVButton").setEnabled(True)

    def _editCSVFile(self):
        src = self.findChild(QLineEdit, "csvFile").text()
        self.__csvEditWindow = CSVEditorWindow(src)
        self.__csvEditWindow.show()



from PyQt6.QtCore import Qt
from PyQt6.QtGui import QFont
from PyQt6.QtWidgets import QLabel, QVBoxLayout, QPushButton
import webbrowser

from niaaml_gui.widgets.base_main_widget import BaseMainWidget


class HelpDocumentationWidget(BaseMainWidget):
    def __init__(self, parent, *args, **kwargs):
        super().__init__(parent, *args, **kwargs)

        vBoxLayout = QVBoxLayout(self._parent)
        vBoxLayout.setAlignment(Qt.AlignmentFlag.AlignTop)

        documentation_label = QLabel("Documentation:")
        font = QFont()
        font.setPointSize(20)
        documentation_label.setFont(font)

        vBoxLayout.addWidget(documentation_label)

        button = QPushButton('Open documentation')
        button.clicked.connect(self.open_link)

        button_font = QFont()
        button_font.setPointSize(12)

        button.setFont(button_font)

        vBoxLayout.addWidget(button)

        self.setLayout(vBoxLayout)

    def open_link(self):
        # Define the URL you want to open
        url = 'https://github.com/firefly-cpp/NiaAML-GUI/blob/master/README.md'

        # Open the URL in the default web browser
        webbrowser.open(url)
